@@ -19,7 +19,7 @@
 /* VTE accessibility object.  Based heavily on inspection of libzvt's
  * accessibility code. */
 
-#ident "$Id: vteaccess.c,v 1.42 2003/06/16 21:23:02 nalin Exp $"
+#ident "$Id: vteaccess.c,v 1.45 2005/08/10 13:29:12 kmaraas Exp $"
 
 #include "../config.h"
 
@@ -62,6 +62,8 @@ enum direction {
 static gunichar vte_terminal_accessible_get_character_at_offset(AtkText *text,
 								gint offset);
 
+static gpointer parent_class = NULL;
+
 /* Create snapshot private data. */
 static VteTerminalAccessiblePrivate *
 vte_terminal_accessible_new_private_data(void)
@@ -103,7 +105,7 @@ vte_terminal_accessible_free_private_data(VteTerminalAccessiblePrivate *priv)
 }
 
 static gint
-offset_from_xy (VteTerminalAccessiblePrivate *priv, 
+offset_from_xy (VteTerminalAccessiblePrivate *priv,
 		gint x, gint y)
 {
 	gint offset;
@@ -116,9 +118,9 @@ offset_from_xy (VteTerminalAccessiblePrivate *priv,
 	linebreak = g_array_index (priv->snapshot_linebreaks, int, y);
 	if (y +1 == priv->snapshot_linebreaks->len)
 		next_linebreak = priv->snapshot_characters->len;
-	else 
+	else
 		next_linebreak = g_array_index (priv->snapshot_linebreaks, int, y + 1);
-	
+
 	offset = linebreak + x;
 	if (offset >= next_linebreak)
 		offset = next_linebreak -1;
@@ -126,7 +128,7 @@ offset_from_xy (VteTerminalAccessiblePrivate *priv,
 }
 
 static void
-xy_from_offset (VteTerminalAccessiblePrivate *priv, 
+xy_from_offset (VteTerminalAccessiblePrivate *priv,
 		gint offset, gint *x, gint *y)
 {
 	gint i;
@@ -140,21 +142,21 @@ xy_from_offset (VteTerminalAccessiblePrivate *priv,
 		linebreak = g_array_index (priv->snapshot_linebreaks, int, i);
 		if (offset < linebreak) {
 			cur_x = offset - cur_offset;
-			cur_y = i - 1;	
+			cur_y = i - 1;
 			break;
-			
+
 		}  else {
 			cur_offset = linebreak;
 		}
 	}
 	if (i == priv->snapshot_linebreaks->len) {
-		if (offset < priv->snapshot_characters->len) {
+		if (offset <= priv->snapshot_characters->len) {
 			cur_x = offset - cur_offset;
-			cur_y = i - 1;	
+			cur_y = i - 1;
 		}
 	}
 	*x = cur_x;
-	*y = cur_y;	
+	*y = cur_y;
 }
 
 /* "Oh yeah, that's selected.  Sure." callback. */
@@ -343,10 +345,10 @@ vte_terminal_accessible_update_private_data_if_needed(AtkObject *text,
 		priv->snapshot_linebreaks = g_array_new(FALSE, TRUE, sizeof(int));
 
 		/* Get a new view of the uber-label. */
-		tmp = vte_terminal_get_text(terminal,
-					    all_selected,
-					    NULL,
-					    priv->snapshot_attributes);
+		tmp = vte_terminal_get_text_include_trailing_spaces(terminal,
+								    all_selected,
+								    NULL,
+								    priv->snapshot_attributes);
 		if (tmp == NULL) {
 			/* Aaargh!  We're screwed. */
 			return;
@@ -732,6 +734,85 @@ vte_terminal_accessible_visibility_notify(VteTerminal *terminal,
 				       visible);
 }
 
+static void
+vte_terminal_accessible_selection_changed (VteTerminal *terminal,
+					   gpointer data)
+{
+	g_return_if_fail(VTE_IS_TERMINAL_ACCESSIBLE(data));
+	g_return_if_fail(VTE_IS_TERMINAL(terminal));
+
+	g_signal_emit_by_name (data, "text_selection_changed");
+}
+
+static void
+vte_terminal_initialize (AtkObject *obj, gpointer data)
+{
+	VteTerminal *terminal;
+	AtkObject *parent;
+
+	ATK_OBJECT_CLASS (parent_class)->initialize (obj, data);
+
+	terminal = VTE_TERMINAL (data);
+
+	_vte_terminal_accessible_ref(terminal);
+
+	g_object_set_data(G_OBJECT(obj),
+			  VTE_TERMINAL_ACCESSIBLE_PRIVATE_DATA,
+			  vte_terminal_accessible_new_private_data());
+
+	g_signal_connect(G_OBJECT(terminal), "text-inserted",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_modified),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "text-deleted",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_modified),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "text-modified",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_modified),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "text-scrolled",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_scrolled),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "cursor-moved",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_invalidate_cursor),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "window-title-changed",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_title_changed),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "focus-in-event",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_focus_in),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "focus-out-event",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_focus_out),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "visibility-notify-event",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_visibility_notify),
+			 obj);
+	g_signal_connect(G_OBJECT(terminal), "selection-changed",
+			 GTK_SIGNAL_FUNC(vte_terminal_accessible_selection_changed),
+			 obj);
+
+	if (GTK_IS_WIDGET((GTK_WIDGET(terminal))->parent)) {
+		parent = gtk_widget_get_accessible((GTK_WIDGET(terminal))->parent);
+		if (ATK_IS_OBJECT(parent)) {
+			atk_object_set_parent(obj, parent);
+		}
+	}
+
+	atk_object_set_name(obj, "Terminal");
+	atk_object_set_description(obj,
+				   terminal->window_title ?
+				   terminal->window_title :
+				   "");
+
+	atk_object_notify_state_change(obj,
+				       ATK_STATE_FOCUSABLE, TRUE);
+	atk_object_notify_state_change(obj,
+				       ATK_STATE_EXPANDABLE, FALSE);
+	atk_object_notify_state_change(obj,
+				       ATK_STATE_RESIZABLE, TRUE);
+	obj->role = ATK_ROLE_TERMINAL;
+}
+
 /**
  * vte_terminal_accessible_new:
  * @terminal: a #VteTerminal
@@ -743,76 +824,16 @@ vte_terminal_accessible_visibility_notify(VteTerminal *terminal,
 AtkObject *
 vte_terminal_accessible_new(VteTerminal *terminal)
 {
-	GtkAccessible *access;
-	AtkObject *parent;
+	AtkObject *accessible;
 	GObject *object;
 
 	g_return_val_if_fail(VTE_IS_TERMINAL(terminal), NULL);
 
 	object = g_object_new(VTE_TYPE_TERMINAL_ACCESSIBLE, NULL);
-	g_return_val_if_fail(GTK_IS_ACCESSIBLE(object), NULL);
+	accessible = ATK_OBJECT (object);
+	atk_object_initialize(accessible, terminal);
 
-	access = GTK_ACCESSIBLE(object);
-	atk_object_initialize(ATK_OBJECT(access), G_OBJECT(terminal));
-
-	access->widget = GTK_WIDGET(terminal);
-	_vte_terminal_accessible_ref(terminal);
-	g_object_add_weak_pointer(G_OBJECT(terminal),
-				  (gpointer*)&access->widget);
-
-	g_object_set_data(G_OBJECT(access),
-			  VTE_TERMINAL_ACCESSIBLE_PRIVATE_DATA,
-			  vte_terminal_accessible_new_private_data());
-
-	g_signal_connect(G_OBJECT(terminal), "text-inserted",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_modified),
-			 object);
-	g_signal_connect(G_OBJECT(terminal), "text-deleted",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_modified),
-			 object);
-	g_signal_connect(G_OBJECT(terminal), "text-modified",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_modified),
-			 object);
-	g_signal_connect(G_OBJECT(terminal), "text-scrolled",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_text_scrolled),
-			 object);
-	g_signal_connect(G_OBJECT(terminal), "cursor-moved",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_invalidate_cursor),
-			 object);
-	g_signal_connect(G_OBJECT(terminal), "window-title-changed",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_title_changed),
-			 access);
-	g_signal_connect(G_OBJECT(terminal), "focus-in-event",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_focus_in),
-			 access);
-	g_signal_connect(G_OBJECT(terminal), "focus-out-event",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_focus_out),
-			 access);
-	g_signal_connect(G_OBJECT(terminal), "visibility-notify-event",
-			 GTK_SIGNAL_FUNC(vte_terminal_accessible_visibility_notify),
-			 access);
-
-	if (GTK_IS_WIDGET((GTK_WIDGET(terminal))->parent)) {
-		parent = gtk_widget_get_accessible((GTK_WIDGET(terminal))->parent);
-		if (ATK_IS_OBJECT(parent)) {
-			atk_object_set_parent(ATK_OBJECT(access), parent);
-		}
-	}
-
-	atk_object_set_name(ATK_OBJECT(access), "Terminal");
-	atk_object_set_description(ATK_OBJECT(access),
-				   terminal->window_title ?
-				   terminal->window_title :
-				   "");
-
-	atk_object_notify_state_change(ATK_OBJECT(access),
-				       ATK_STATE_FOCUSABLE, TRUE);
-	atk_object_notify_state_change(ATK_OBJECT(access),
-				       ATK_STATE_EXPANDABLE, FALSE);
-	atk_object_notify_state_change(ATK_OBJECT(access),
-				       ATK_STATE_RESIZABLE, TRUE);
-
-	return ATK_OBJECT(access);
+	return accessible;
 }
 
 static void
@@ -1231,19 +1252,19 @@ get_attribute_set (struct _VteCharAttributes attr)
 	}
 	at = g_new (AtkAttribute, 1);
 	at->name = g_strdup ("fg-color");
-	at->value = g_strdup_printf ("%u,%u,%u", 
+	at->value = g_strdup_printf ("%u,%u,%u",
 				     attr.fore.red, attr.fore.green, attr.fore.blue);
 	set = g_slist_append (set, at);
 
 	at = g_new (AtkAttribute, 1);
 	at->name = g_strdup ("bg-color");
-	at->value = g_strdup_printf ("%u,%u,%u", 
+	at->value = g_strdup_printf ("%u,%u,%u",
 				     attr.back.red, attr.back.green, attr.back.blue);
 	set = g_slist_append (set, at);
 
 	return set;
 }
-	
+
 static AtkAttributeSet *
 vte_terminal_accessible_get_run_attributes(AtkText *text, gint offset,
 					   gint *start_offset, gint *end_offset)
@@ -1271,7 +1292,7 @@ vte_terminal_accessible_get_run_attributes(AtkText *text, gint offset,
 		    !gdk_color_equal (&cur_attr.back, &attr.back) ||
 		    cur_attr.underline != attr.underline ||
 		    cur_attr.strikethrough != attr.strikethrough) {
-			*start_offset = i + 1;	
+			*start_offset = i + 1;
 			break;
 		}
 	}
@@ -1288,7 +1309,7 @@ vte_terminal_accessible_get_run_attributes(AtkText *text, gint offset,
 			break;
 		}
 	}
-	
+
 	return get_attribute_set (attr);
 }
 
@@ -1374,55 +1395,136 @@ vte_terminal_accessible_get_offset_at_point(AtkText *text,
 static gint
 vte_terminal_accessible_get_n_selections(AtkText *text)
 {
-	g_return_val_if_fail(VTE_IS_TERMINAL_ACCESSIBLE(text), 0);
+	GtkWidget *widget;
+	VteTerminal *terminal;
+
+	g_return_val_if_fail(VTE_IS_TERMINAL_ACCESSIBLE(text), -1);
 	vte_terminal_accessible_update_private_data_if_needed(ATK_OBJECT(text),
 							      NULL, NULL);
-	/* FIXME? */
-	return 0;
+
+	widget = GTK_ACCESSIBLE(text)->widget;
+	if (widget == NULL) {
+		/* State is defunct */
+		return -1;
+	}
+	g_return_val_if_fail (VTE_IS_TERMINAL (widget), -1);
+	terminal = VTE_TERMINAL (widget);
+	return (vte_terminal_get_has_selection (terminal)) ? 1 : 0;
 }
 
 static gchar *
 vte_terminal_accessible_get_selection(AtkText *text, gint selection_number,
 				      gint *start_offset, gint *end_offset)
 {
+	GtkWidget *widget;
+	VteTerminal *terminal;
+	VteTerminalAccessiblePrivate *priv;
+	long start_x, start_y, end_x, end_y;
+
 	g_return_val_if_fail(VTE_IS_TERMINAL_ACCESSIBLE(text), NULL);
 	vte_terminal_accessible_update_private_data_if_needed(ATK_OBJECT(text),
 							      NULL, NULL);
-	/* FIXME? */
-	return NULL;
+	widget = GTK_ACCESSIBLE(text)->widget;
+	if (widget == NULL) {
+		/* State is defunct */
+		return NULL;
+	}
+	g_return_val_if_fail (VTE_IS_TERMINAL (widget), NULL);
+	terminal = VTE_TERMINAL (widget);
+	if (!vte_terminal_get_has_selection (terminal)) {
+		return NULL;
+	}
+	if (selection_number != 0) {
+		return NULL;
+	}
+
+	priv = g_object_get_data(G_OBJECT(text),
+				 VTE_TERMINAL_ACCESSIBLE_PRIVATE_DATA);
+	_vte_terminal_get_start_selection (terminal, &start_x, &start_y);
+	*start_offset = offset_from_xy (priv, start_x, start_y);
+	_vte_terminal_get_end_selection (terminal, &end_x, &end_y);
+	*end_offset = offset_from_xy (priv, end_x, end_y);
+	return _vte_terminal_get_selection (terminal);
 }
 
 static gboolean
 vte_terminal_accessible_add_selection(AtkText *text,
 				      gint start_offset, gint end_offset)
 {
+	GtkWidget *widget;
+	VteTerminal *terminal;
+	VteTerminalAccessiblePrivate *priv;
+	gint start_x, start_y, end_x, end_y;
+
 	g_return_val_if_fail(VTE_IS_TERMINAL_ACCESSIBLE(text), FALSE);
 	vte_terminal_accessible_update_private_data_if_needed(ATK_OBJECT(text),
 							      NULL, NULL);
-	/* FIXME? */
-	return FALSE;
+	widget = GTK_ACCESSIBLE(text)->widget;
+	if (widget == NULL) {
+		/* State is defunct */
+		return FALSE;
+	}
+	g_return_val_if_fail (VTE_IS_TERMINAL (widget), FALSE);
+	terminal = VTE_TERMINAL (widget);
+	g_return_val_if_fail (!vte_terminal_get_has_selection (terminal), FALSE);
+	priv = g_object_get_data(G_OBJECT(text),
+				 VTE_TERMINAL_ACCESSIBLE_PRIVATE_DATA);
+	xy_from_offset (priv, start_offset, &start_x, &start_y);
+	xy_from_offset (priv, end_offset, &end_x, &end_y);
+	_vte_terminal_select_text (terminal, start_x, start_y, end_x, end_y, start_offset, end_offset);
+	return TRUE;
 }
 
 static gboolean
 vte_terminal_accessible_remove_selection(AtkText *text,
 					 gint selection_number)
 {
+	GtkWidget *widget;
+	VteTerminal *terminal;
+
 	g_return_val_if_fail(VTE_IS_TERMINAL_ACCESSIBLE(text), FALSE);
 	vte_terminal_accessible_update_private_data_if_needed(ATK_OBJECT(text),
 							      NULL, NULL);
-	/* FIXME? */
-	return FALSE;
+	widget = GTK_ACCESSIBLE(text)->widget;
+	if (widget == NULL) {
+		/* State is defunct */
+		return FALSE;
+	}
+	g_return_val_if_fail (VTE_IS_TERMINAL (widget), FALSE);
+	terminal = VTE_TERMINAL (widget);
+	if (selection_number == 0 && vte_terminal_get_has_selection (terminal)) {
+		_vte_terminal_remove_selection (terminal);
+		return TRUE;
+	} else {
+		return FALSE;
+	}
 }
 
 static gboolean
 vte_terminal_accessible_set_selection(AtkText *text, gint selection_number,
 				      gint start_offset, gint end_offset)
 {
+	GtkWidget *widget;
+	VteTerminal *terminal;
+
 	g_return_val_if_fail(VTE_IS_TERMINAL_ACCESSIBLE(text), FALSE);
 	vte_terminal_accessible_update_private_data_if_needed(ATK_OBJECT(text),
 							      NULL, NULL);
-	/* FIXME? */
-	return FALSE;
+	widget = GTK_ACCESSIBLE(text)->widget;
+	if (widget == NULL) {
+		/* State is defunct */
+		return FALSE;
+	}
+	g_return_val_if_fail (VTE_IS_TERMINAL (widget), FALSE);
+	terminal = VTE_TERMINAL (widget);
+	if (selection_number != 0) {
+		return FALSE;
+	}
+	if (vte_terminal_get_has_selection (terminal)) {
+		_vte_terminal_remove_selection (terminal);
+	}
+
+	return vte_terminal_accessible_add_selection (text, start_offset, end_offset);
 }
 
 static gboolean
@@ -1511,6 +1613,9 @@ vte_terminal_accessible_get_position(AtkComponent *component,
 	*x = 0;
 	*y = 0;
 	widget = (GTK_ACCESSIBLE(component))->widget;
+	if (widget == NULL) {
+		return;
+	}
 	if (!GTK_WIDGET_REALIZED(widget)) {
 		return;
 	}
@@ -1535,6 +1640,9 @@ vte_terminal_accessible_get_size(AtkComponent *component,
 	*width = 0;
 	*height = 0;
 	widget = (GTK_ACCESSIBLE(component))->widget;
+	if (widget == NULL) {
+		return;
+	}
 	if (!GTK_WIDGET_REALIZED(widget)) {
 		return;
 	}
@@ -1567,7 +1675,12 @@ vte_terminal_accessible_set_size(AtkComponent *component,
 {
 	VteTerminal *terminal;
 	gint columns, rows, xpad, ypad;
-	terminal = VTE_TERMINAL((GTK_ACCESSIBLE(component))->widget);
+	GtkWidget *widget;
+	widget = GTK_ACCESSIBLE(component)->widget;
+	if (widget == NULL) {
+		return FALSE;
+	}
+	terminal = VTE_TERMINAL(widget);
 	vte_terminal_get_padding(terminal, &xpad, &ypad);
 	/* If the size is an exact multiple of the cell size, use that,
 	 * otherwise round down. */
@@ -1591,6 +1704,9 @@ vte_terminal_accessible_grab_focus(AtkComponent *component)
 {
 	GtkWidget *widget;
 	widget = (GTK_ACCESSIBLE(component))->widget;
+	if (widget == NULL) {
+		return FALSE;
+	}
 	if (GTK_WIDGET_HAS_FOCUS(widget)) {
 		return TRUE;
 	}
@@ -1671,21 +1787,17 @@ static void
 vte_terminal_accessible_class_init(gpointer *klass)
 {
 	GObjectClass *gobject_class;
+	AtkObjectClass *class = ATK_OBJECT_CLASS (klass);
 
 	gobject_class = G_OBJECT_CLASS(klass);
 
+	parent_class = g_type_class_peek_parent (klass);
+
+	class->initialize = vte_terminal_initialize;
 	/* Override the finalize method. */
 	gobject_class->finalize = vte_terminal_accessible_finalize;
 }
 
-static void
-vte_terminal_accessible_init(gpointer *instance, gpointer *klass)
-{
-	/* Mark the role this object plays. */
-	g_return_if_fail(ATK_IS_OBJECT(instance));
-	atk_object_set_role(ATK_OBJECT(instance), ATK_ROLE_TERMINAL);
-}
-
 GType
 vte_terminal_accessible_get_type(void)
 {
@@ -1714,9 +1826,9 @@ vte_terminal_accessible_get_type(void)
 		(GClassFinalizeFunc)NULL,
 		(gconstpointer)NULL,
 
-		sizeof(VteTerminalAccessible),
 		0,
-		(GInstanceInitFunc)vte_terminal_accessible_init,
+		0,
+		(GInstanceInitFunc) NULL,
 
 		(GTypeValueTable*)NULL,
 	};
